import importlib
import pkgutil
from types import ModuleType
from typing import Dict

import pkg_resources


def all_installed_packages() -> Dict[str, str]:
    return {
        entry.project_name: entry.version
        # working_set is actually iterable see sources
        for entry in pkg_resources.working_set  # pylint: disable=not-an-iterable
    }


cached_installed_packages = all_installed_packages()


def load_all_modules_from(module: ModuleType) -> None:
    for loader, name, is_pkg in pkgutil.walk_packages(module.__path__):
        importlib.import_module(module.__name__ + '.' + name)



import abc
import dataclasses
import json
import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, BinaryIO, Callable, Dict, Optional, Type, Union

from packaging import version  # type: ignore

_LOG = logging.getLogger(__name__)


class StandardDataFormats(Enum):
    pickle = "pickle"
    proto = "proto"
    primitive_type = "primitive_type"


class StandardSchemaFormats(Enum):
    pickled_type = "pickled_type"
    no_schema = "no_schema"


@dataclass(frozen=True)
class Schema:
    data_format: str
    schema_format: str
    schema_content: str = ''
    meta: Dict[str, str] = field(default_factory=lambda: {})


@dataclass(frozen=True)
class VersionBoundary:
    greater_than: Optional[str] = None
    less_than: Optional[str] = None


class Serializer(abc.ABC):
    HEADER_BYTES = 'serialzy'.encode('utf-8')
    HEADER_BYTES_LEN = len(HEADER_BYTES)

    def serialize(self, obj: Any, dest: BinaryIO) -> None:
        """
        :param obj: object to serialize into bytes
        :param dest: serialized obj is written into dest
        :return: None
        """

        typ = type(obj)
        # check that obj type is valid
        self._check_type(typ)
        # write schema header
        self._write_schema(typ, dest)
        # write serialized data
        self._serialize(obj, dest)

    @abc.abstractmethod
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        """
        :param obj: object to serialize into bytes
        :param dest: serialized obj is written into dest
        :return: None
        """

    def deserialize(self, source: BinaryIO, typ: Optional[Type] = None) -> Any:
        """
        :param source: buffer of file with serialized data
        :param typ: type of the resulting object, fetched from the header if None
        :return: deserialized object
        """

        # read schema header
        schema_type = self._deserialize_type(source)

        # read serialized data
        return self._deserialize(source, schema_type, typ)

    @abc.abstractmethod
    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        """
        :param source: buffer of file with serialized data
        :param schema_type: type of the resulting object retrieved from the schema header
        :param user_type: type of the resulting object provided by a user
        :return: deserialized object
        """

    @abc.abstractmethod
    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        """
        :return: type suitable for the serializer or types filter
        """

    @abc.abstractmethod
    def available(self) -> bool:
        """
        :return: True if the serializer can be used in the current environment, otherwise False
        """

    @abc.abstractmethod
    def stable(self) -> bool:
        """
        :return: True if the serializer does not depend on python version/dependency versions/etc., otherwise False
        """

    @abc.abstractmethod
    def data_format(self) -> str:
        """
        :return: data format that this serializer is working with
        """

    @abc.abstractmethod
    def meta(self) -> Dict[str, str]:
        """
        :return: meta of this serializer, e.g., versions of dependencies
        """

    @abc.abstractmethod
    def schema(self, typ: type) -> Schema:
        """
        :param typ: type of object for serialization
        :return: schema for the object
        """

    @abc.abstractmethod
    def resolve(self, schema: Schema) -> Type:
        """
        :param schema: schema that contains information about serialized data
        :return: Type used for python representation of the schema
        """

    @abc.abstractmethod
    def requirements(self) -> Dict[str, VersionBoundary]:
        """
        :return: the requirements (libraries) which are needed to be installed to make serializer available
        """

    def _validate_schema(self, schema: Schema) -> None:
        if schema.data_format != self.data_format():
            raise ValueError(
                f"Invalid data format {schema.data_format}, expected {self.data_format()}"
            )

    def _deserialize_type(self, source: BinaryIO) -> Type:
        first_str = source.read(self.HEADER_BYTES_LEN)
        if first_str != self.HEADER_BYTES:
            raise ValueError('Invalid source format')

        header_len = int.from_bytes(source.read(8), byteorder='little', signed=False)
        schema_json = source.read(header_len).decode('utf-8')
        schema = Schema(**json.loads(schema_json))
        typ = self.resolve(schema)
        return typ

    def _write_schema(self, typ: Type, dest: BinaryIO) -> None:
        schema = self.schema(typ)
        schema_json = json.dumps(dataclasses.asdict(schema))
        schema_bytes = schema_json.encode('utf-8')
        header_len = len(schema_bytes)
        dest.write(self.HEADER_BYTES)
        dest.write(header_len.to_bytes(length=8, byteorder='little', signed=False))
        dest.write(schema_bytes)

    def _check_type(self, typ: Type) -> None:
        supported = self.supported_types()
        # mypy issue: https://github.com/python/mypy/issues/3060
        if (isinstance(supported, Type) and typ != supported) or (  # type: ignore
                not isinstance(supported, Type) and not supported(typ)):  # type: ignore
            raise ValueError(f'Invalid object type {typ} for the serializer {type(self)}')

    @staticmethod
    def _check_types_valid(schema_type: Type, user_type: Optional[Type]) -> None:
        if user_type is not None and user_type != schema_type:
            raise ValueError(f'Cannot deserialize data with schema type {schema_type} into type {user_type}')


class SerializerRegistry(abc.ABC):
    @abc.abstractmethod
    def register_serializer(self, serializer: Serializer, priority: Optional[int] = None) -> None:
        """
        :param serializer: serializer to register
        :param priority: number that indicates serializer's priority: 0 - max priority
        :return: None
        """

    @abc.abstractmethod
    def unregister_serializer(self, serializer: Serializer) -> None:
        """
        :param serializer: serializer to unregister
        :return:
        """

    @abc.abstractmethod
    def find_serializer_by_type(self, typ: Type) -> Optional[Serializer]:
        """
        :param typ: python Type needed to serialize
        :return: corresponding serializer or None
        """

    @abc.abstractmethod
    def find_serializer_by_data_format(self, data_format: str) -> Optional[Serializer]:
        """
        :param data_format: data format to resolve serializer
        :return: Serializer if there is a serializer for that data format, None otherwise
        """

    @abc.abstractmethod
    def reload_registry(self) -> None:
        """
        reloads all registered serializers, useful if libraries are updated
        """



import inspect
import logging
import sys
from collections import defaultdict
from types import ModuleType
from typing import Dict, List, Optional, Type, cast, Iterable

import serialzy.serializers
from serialzy.api import Serializer, SerializerRegistry
from serialzy.cloudpickle import CloudpickleSerializer
from serialzy.utils import load_all_modules_from

_LOG = logging.getLogger(__name__)


class DefaultSerializerRegistry(SerializerRegistry):
    def __init__(self):
        self._default_priority_stable = sys.maxsize - 10
        self._default_priority_unstable = sys.maxsize - 5

        self._type_registry: Dict[Type, List[Serializer]] = defaultdict(list)
        self._data_formats_serializer_registry: Dict[str, List[Serializer]] = defaultdict(list)

        self._serializer_priorities: Dict[Type[Serializer], int] = {}
        self._serializer_registry: Dict[Type[Serializer], Serializer] = {}

        load_all_modules_from(serialzy.serializers)
        for serializer in self._fetch_serializers_from(serialzy.serializers):
            if serializer.stable():
                self.register_serializer(serializer, self._default_priority_stable)
            else:
                self.register_serializer(serializer, self._default_priority_unstable)
        # cloudpickle has the least priority
        self.register_serializer(CloudpickleSerializer(), sys.maxsize - 1)

    def register_serializer(self, serializer: Serializer, priority: Optional[int] = None) -> None:
        serializer_type = type(serializer)
        if serializer_type in self._serializer_registry:
            raise ValueError(f"Serializer {serializer_type} has been already registered")

        self._serializer_registry[serializer_type] = serializer
        self._data_formats_serializer_registry[serializer.data_format()].append(serializer)

        if priority is None:
            if serializer.stable():
                priority = self._default_priority_stable
            else:
                priority = self._default_priority_unstable
        self._serializer_priorities[serializer_type] = priority

        try:
            # mypy issue: https://github.com/python/mypy/issues/3060
            if isinstance(serializer.supported_types(), Type):  # type: ignore
                self._type_registry[cast(Type, serializer.supported_types())].append(serializer)
        except (ImportError, ModuleNotFoundError):
            pass

    def unregister_serializer(self, serializer: Serializer):
        serializer_type = type(serializer)
        if serializer_type in self._serializer_registry:
            self._data_formats_serializer_registry[serializer.data_format()].remove(serializer)
            del self._serializer_priorities[serializer_type]
            del self._serializer_registry[serializer_type]

            try:
                # mypy issue: https://github.com/python/mypy/issues/3060
                if isinstance(serializer.supported_types(),  # type: ignore
                              Type) and serializer.supported_types() in self._type_registry:  # type: ignore
                    del self._type_registry[cast(Type, serializer.supported_types())]
            except (ImportError, ModuleNotFoundError):
                pass

    def find_serializer_by_type(self, typ: Type) -> Optional[Serializer]:
        result: Optional[Serializer] = None
        priority = sys.maxsize
        for serializer_type, serializer in self._serializer_registry.items():
            try:
                if (
                        # mypy issue: https://github.com/python/mypy/issues/3060
                        not isinstance(serializer.supported_types(), Type)  # type: ignore
                        and serializer.supported_types()(typ)
                        and self._serializer_priorities[serializer_type] < priority
                ):
                    priority = self._serializer_priorities[serializer_type]
                    result = serializer
            except (ImportError, ModuleNotFoundError):
                continue

        for serializer in self._type_registry[typ]:
            if self._serializer_priorities[type(serializer)] <= priority:
                result = serializer

        return result

    def find_serializer_by_data_format(self, data_format: str) -> Optional[Serializer]:
        serializer: Optional[Serializer] = None
        serializer_priority = sys.maxsize
        for ser in self._data_formats_serializer_registry[data_format]:
            ser_type = type(ser)
            if self._serializer_priorities[ser_type] < serializer_priority:
                serializer = ser
                serializer_priority = self._serializer_priorities[ser_type]
        return serializer

    def reload_registry(self) -> None:
        for typ, serializer in self._serializer_registry.copy().items():
            priority = self._serializer_priorities[typ]
            self.unregister_serializer(serializer)
            self.register_serializer(serializer, priority)

    def _fetch_serializers_from(self, module: ModuleType) -> Iterable[Serializer]:
        stable_serializer_modules = dir(module)
        for module_attr in stable_serializer_modules:
            module_value = getattr(module, module_attr)
            if isinstance(module_value, ModuleType):
                stable_serializers = dir(module_value)
                for class_attr in stable_serializers:
                    class_value = getattr(module_value, class_attr)
                    if inspect.isclass(class_value) and not inspect.isabstract(class_value) and issubclass(class_value,
                                                                                                           Serializer):
                        sig = inspect.signature(class_value)
                        if len(sig.parameters) == 0:
                            instance = class_value()
                        elif len(sig.parameters) == 1 and issubclass(list(sig.parameters.values())[0].annotation,
                                                                     SerializerRegistry):
                            # noinspection PyArgumentList
                            instance = class_value(self)  # type: ignore
                        else:
                            raise ValueError(
                                f'Serializer {class_value.__name__} has unexpected arguments in __init__: '
                                f'only empty arguments or SerializerRegistry are allowed')

                        yield instance



import logging
import inspect
from typing import Any, BinaryIO, Callable, Dict, Type, Union, Optional

from serialzy.api import StandardDataFormats, Schema, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByValue
from serialzy.utils import cached_installed_packages
from packaging import version  # type: ignore

_LOG = logging.getLogger(__name__)


# noinspection PyMethodMayBeStatic
class ProtoMessageSerializer(DefaultSchemaSerializerByValue):
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        obj.dump(dest)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        from pure_protobuf.dataclasses_ import load, Message  # type: ignore

        if user_type is not None:
            if not issubclass(user_type, Message):
                raise ValueError(f'Cannot deserialize data with schema type {schema_type} into type {user_type}')
            # noinspection PyTypeChecker
            return load(user_type, source)

        # noinspection PyTypeChecker
        return load(schema_type, source)

    def available(self) -> bool:
        base_available = super().available()
        if not base_available:
            return False

        # noinspection PyBroadException
        try:
            import pure_protobuf  # type: ignore
            return True
        except:
            return False

    def stable(self) -> bool:
        return True

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        from pure_protobuf.dataclasses_ import Message  # type: ignore

        return lambda t: inspect.isclass(t) and issubclass(t, Message)

    def data_format(self) -> str:
        return StandardDataFormats.proto.name

    def meta(self) -> Dict[str, str]:
        return {"pure-protobuf": cached_installed_packages["pure-protobuf"]}

    def resolve(self, schema: Schema) -> Type:
        typ = super().resolve(schema)
        if 'pure-protobuf' not in schema.meta:
            _LOG.warning('No pure-protobuf version in meta')
        elif version.parse(schema.meta['pure-protobuf']) > version.parse(cached_installed_packages["pure-protobuf"]):
            _LOG.warning(f'Installed version of pure-protobuf {cached_installed_packages["pure-protobuf"]} '
                         f'is older than used for serialization {schema.meta["pure-protobuf"]}')
        return typ

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {'pure-protobuf': VersionBoundary()}



import logging
from typing import Any, BinaryIO, Callable, Dict, Type, Union, Optional

from packaging import version  # type: ignore

from serialzy.api import StandardDataFormats, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByReference
from serialzy.version import __version__

_LOG = logging.getLogger(__name__)


# noinspection PyPackageRequirements
class PrimitiveSerializer(DefaultSchemaSerializerByReference):
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        dumps = str(obj).encode("utf-8")
        dest.write(dumps)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        read = source.read().decode("utf-8")
        if schema_type == type(None):
            return None
        return schema_type(read)

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: t in [int, float, str, bool, type(None)]

    def available(self) -> bool:
        return True

    def stable(self) -> bool:
        return True

    def data_format(self) -> str:
        return StandardDataFormats.primitive_type.name

    def meta(self) -> Dict[str, str]:
        return {'serialzy': __version__}

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {}



import dataclasses
import json
import logging
import tempfile
from abc import ABC
from typing import Type, Dict, Union, BinaryIO, Any, cast, Optional, Callable, Tuple

from packaging import version  # type: ignore
from typing_extensions import get_args, get_origin

from serialzy.api import Serializer, Schema, SerializerRegistry, VersionBoundary
from serialzy.utils import cached_installed_packages
from serialzy.version import __version__

_LOG = logging.getLogger(__name__)


class UnionSerializerBase(Serializer, ABC):
    SCHEMA_FORMAT = "serialzy_union_schema"

    def __init__(self, registry: SerializerRegistry):
        self._registry = registry

    def serialize(self, obj: Any, dest: BinaryIO) -> None:
        typ = type(obj)
        serializer = self._registry.find_serializer_by_type(typ)
        if serializer is None:
            raise ValueError(f'Cannot find serializer for type {typ}')
        serializer.serialize(obj, dest)

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        """
        Don't use this method for union serialization
        """

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        if user_type is None or get_origin(user_type) != Union:
            serializer = cast(Serializer, self._registry.find_serializer_by_type(schema_type))
            return serializer._deserialize(source, schema_type, user_type)
        else:
            args = get_args(user_type)
            # Optional case
            if len(args) == 2 and args[1] == type(None):
                if schema_type == type(None):
                    return cast(Serializer, self._registry.find_serializer_by_type(schema_type))._deserialize(source,
                                                                                                              schema_type)
                else:
                    try:
                        return cast(Serializer, self._registry.find_serializer_by_type(args[0]))._deserialize(source,
                                                                                                              schema_type,
                                                                                                              args[0])
                    except:
                        raise ValueError(f'Cannot deserialize data into type {user_type}')
            # General union case
            else:
                with tempfile.NamedTemporaryFile('wb+') as handle:
                    # copy source
                    while True:
                        data = source.read(8096)
                        if not data:
                            break
                        handle.write(data)

                    handle.flush()
                    handle.seek(0)

                    # try to deserialize in union arguments
                    for arg in args:
                        # noinspection PyBroadException
                        try:
                            serializer_by_type = cast(Serializer, self._registry.find_serializer_by_type(arg))
                            obj = serializer_by_type._deserialize(cast(BinaryIO, handle), schema_type, arg)
                            return obj
                        except:
                            handle.seek(0)
                            continue

                    raise ValueError(f'Cannot deserialize data into type {user_type}')

    def available(self) -> bool:
        return True

    def data_format(self) -> str:
        return "serialzy_union"

    def meta(self) -> Dict[str, str]:
        return {'serialzy': __version__}

    def schema(self, typ: type) -> Schema:
        args = get_args(typ)
        schemas = [dataclasses.asdict(cast(Serializer, self._registry.find_serializer_by_type(x)).schema(x))
                   for x in args]
        return Schema(self.data_format(), self.SCHEMA_FORMAT, json.dumps(schemas), self.meta())

    def resolve(self, schema: Schema) -> Type:
        if schema.data_format == self.data_format():
            self._validate_schema(schema)
            if schema.schema_format != self.SCHEMA_FORMAT:
                raise ValueError(f'Invalid schema format {schema.schema_format}')

            if 'serialzy' not in schema.meta:
                _LOG.warning('No serialzy version in meta')
            elif version.parse(schema.meta['serialzy']) > version.parse(cached_installed_packages["serialzy"]):
                _LOG.warning(f'Installed version of serialzy {cached_installed_packages["serialzy"]} '
                             f'is older than used for serialization {schema.meta["serialzy"]}')

            schemas = json.loads(schema.schema_content)
            types = tuple(
                cast(Serializer, self._registry.find_serializer_by_data_format(Schema(**x).data_format))
                .resolve(Schema(**x)) for x in schemas
            )
            return cast(Type, Union[types])

        serializer = self._registry.find_serializer_by_data_format(schema.data_format)
        if serializer is None:
            raise ValueError(f'Cannot find serializer for data format {schema.data_format}')
        return serializer.resolve(schema)

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {}


class UnionSerializerStable(UnionSerializerBase):
    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: get_origin(t) == Union and self.__check_args(get_args(t))

    def stable(self) -> bool:
        return True

    def __check_args(self, args: Tuple[Any, ...]) -> bool:
        for arg in args:
            serializer = self._registry.find_serializer_by_type(arg)
            if serializer is None or not serializer.available() or not serializer.stable():
                return False
        return True


class UnionSerializerUnstable(UnionSerializerBase):
    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: get_origin(t) == Union and self.__check_args(get_args(t))

    def stable(self) -> bool:
        return False

    def __check_args(self, args: Tuple[Any, ...]) -> bool:
        for arg in args:
            serializer = self._registry.find_serializer_by_type(arg)
            if serializer is None or not serializer.available():
                return False
        return True



import logging
from typing import Any, BinaryIO, Callable, Dict, Type, Union, Optional

from serialzy.api import StandardDataFormats, Schema, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByValue
from packaging import version  # type: ignore

from serialzy.utils import cached_installed_packages

_LOG = logging.getLogger(__name__)


# noinspection PyMethodMayBeStatic, PyPackageRequirements
class CloudpickleSerializer(DefaultSchemaSerializerByValue):
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import cloudpickle  # type: ignore
        cloudpickle.dump(obj, dest)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import cloudpickle  # type: ignore
        return cloudpickle.load(source)

    def available(self) -> bool:
        base_available = super().available()
        if not base_available:
            return False
        # noinspection PyBroadException
        try:
            import cloudpickle  # type: ignore
            return True
        except:
            return False

    def stable(self) -> bool:
        return False

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda x: True

    def data_format(self) -> str:
        return StandardDataFormats.pickle.name

    def meta(self) -> Dict[str, str]:
        import cloudpickle  # type: ignore
        return {"cloudpickle": cloudpickle.__version__}

    def resolve(self, schema: Schema) -> Type:
        typ = super().resolve(schema)
        if 'cloudpickle' not in schema.meta:
            _LOG.warning('No cloudpickle version in meta')
        elif version.parse(schema.meta['cloudpickle']) > version.parse(cached_installed_packages["cloudpickle"]):
            _LOG.warning(f'Installed version of cloudpickle {cached_installed_packages["cloudpickle"]} '
                         f'is older than used for serialization {schema.meta["cloudpickle"]}')
        return typ

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {"cloudpickle": VersionBoundary()}



import logging
import os
import tempfile
from abc import ABC
from typing import BinaryIO, Callable, Dict, Type, Union, Any, Optional

from packaging import version  # type: ignore

from serialzy.api import Schema, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByReference
from serialzy.errors import SerialzyError
from serialzy.utils import cached_installed_packages

_LOG = logging.getLogger(__name__)


# noinspection PyPackageRequirements
class CatboostBaseSerializer(DefaultSchemaSerializerByReference, ABC):
    def available(self) -> bool:
        # noinspection PyBroadException
        try:
            import catboost  # type: ignore

            return True
        except:
            return False

    def stable(self) -> bool:
        return True

    def meta(self) -> Dict[str, str]:
        import catboost

        return {"catboost": catboost.__version__}

    def resolve(self, schema: Schema) -> Type:
        typ = super().resolve(schema)
        if 'catboost' not in schema.meta:
            _LOG.warning('No catboost version in meta')
        elif version.parse(schema.meta['catboost']) > version.parse(cached_installed_packages["catboost"]):
            _LOG.warning(f'Installed version of catboost {cached_installed_packages["catboost"]} '
                         f'is older than used for serialization {schema.meta["catboost"]}')
        return typ

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {'catboost': VersionBoundary()}


# noinspection PyPackageRequirements
class CatboostPoolSerializer(CatboostBaseSerializer):
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        with tempfile.NamedTemporaryFile() as handle:
            if not obj.is_quantized():  # type: ignore
                raise SerialzyError('Only quantized pools can be serialized')
            obj.save(handle.name)  # type: ignore
            while True:
                data = handle.read(8096)
                if not data:
                    break
                dest.write(data)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        with tempfile.NamedTemporaryFile() as handle:
            while True:
                data = source.read(8096)
                if not data:
                    break
                handle.write(data)
            handle.flush()
            os.fsync(handle.fileno())
            import catboost

            return catboost.Pool("quantized://" + handle.name)  # type: ignore

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import catboost

        return catboost.Pool  # type: ignore

    def data_format(self) -> str:
        return "catboost_quantized_pool"


# noinspection PyPackageRequirements
class CatboostModelSerializer(CatboostBaseSerializer):
    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        with tempfile.NamedTemporaryFile() as handle:
            obj.save_model(handle.name, format=self.data_format())  # type: ignore
            while True:
                data = handle.read(8096)
                if not data:
                    break
                dest.write(data)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        with tempfile.NamedTemporaryFile() as handle:
            while True:
                data = source.read(8096)
                if not data:
                    break
                handle.write(data)
            handle.flush()
            os.fsync(handle.fileno())

            model = schema_type()
            model.load_model(handle.name, format=self.data_format())  # type: ignore
            return model

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import catboost as cb
        return lambda t: t in [cb.CatBoostClassifier, cb.CatBoostRegressor, cb.CatBoostRanker]  # type: ignore

    def data_format(self) -> str:
        return "cbm"  # CatBoost binary format https://catboost.ai/en/docs/concepts/python-reference_catboost_save_model



import base64
import importlib
import json
import logging
from abc import ABC
from typing import Type, cast
from packaging import version  # type: ignore
from serialzy.utils import cached_installed_packages

from serialzy.api import Serializer, Schema, StandardSchemaFormats

_LOG = logging.getLogger(__name__)


# noinspection PyPackageRequirements
class DefaultSchemaSerializerByValue(Serializer, ABC):
    def available(self) -> bool:
        # noinspection PyBroadException
        try:
            import cloudpickle  # type: ignore
            return True
        except:
            return False

    def schema(self, typ: type) -> Schema:
        import cloudpickle  # type: ignore
        return Schema(
            self.data_format(),
            StandardSchemaFormats.pickled_type.name,
            base64.b64encode(cloudpickle.dumps(typ)).decode("ascii"),
            {**self.meta(), **{"cloudpickle": cloudpickle.__version__}},
        )

    def resolve(self, schema: Schema) -> Type:
        import cloudpickle  # type: ignore
        self._validate_schema(schema)
        if schema.schema_format != StandardSchemaFormats.pickled_type.name:
            raise ValueError(f'Invalid schema format: {schema.schema_format}')
        if 'cloudpickle' not in schema.meta:
            _LOG.warning('No cloudpickle version in meta')
        elif version.parse(schema.meta['cloudpickle']) > version.parse(cached_installed_packages["cloudpickle"]):
            _LOG.warning(f'Installed version of cloudpickle {cached_installed_packages["cloudpickle"]} '
                         f'is older than used for serialization {schema.meta["cloudpickle"]}')
        return cast(
            Type,
            cloudpickle.loads(
                base64.b64decode(cast(str, schema.schema_content).encode("ascii"))
            ),
        )


class DefaultSchemaSerializerByReference(Serializer, ABC):
    SCHEMA_FORMAT = "serialzy_python_type_reference"

    def schema(self, typ: type) -> Schema:
        return Schema(
            self.data_format(),
            self.SCHEMA_FORMAT,
            json.dumps({
                "module": typ.__module__,
                "name": typ.__name__
            }),
            self.meta(),
        )

    def resolve(self, schema: Schema) -> Type:
        self._validate_schema(schema)
        if schema.schema_format != self.SCHEMA_FORMAT:
            raise ValueError(f'Invalid schema format: {schema.schema_content}')
        info = json.loads(schema.schema_content)
        if 'module' not in info or 'name' not in info:
            raise ValueError(f'Invalid schema content: {schema.schema_content}')

        module = info['module']
        name = info['name']
        # NoneType cannot be imported from builtin
        if module == 'builtins' and name == 'NoneType':
            return type(None)

        typ = getattr(importlib.import_module(module), name)
        return cast(type, typ)



from typing import List

import setuptools


def read_version(path="serialzy/version/version"):
    with open(path) as file:
        return file.read().rstrip()


def read_requirements() -> List[str]:
    requirements = []
    with open("requirements.txt", "r") as file:
        for line in file:
            requirements.append(line.rstrip())
    return requirements


def read_readme() -> str:
    with open("README.md", "r") as file:
        return file.read()


setuptools.setup(
    name="serialzy",
    version=read_version(),
    license="LICENSE",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],
    author="ʎzy developers",
    install_requires=read_requirements(),
    package_data={
        "serialzy": ["version/version"],
    },
    packages=[
        "serialzy",
        "serialzy/serializers"
    ],
    python_requires=">=3.7",
    long_description=read_readme(),
    long_description_content_type='text/markdown'
)
